unit RTCP;

// RTCP component for Delphi (c)2006 by Paul TOTH <tothpaul@free.fr>
// http://tothpaul.free.fr

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

{-$define LISTEN} // should we listen for RTCP packets ?

interface

uses Windows,Winsock,UDP,RTP;

type
{
        0                   1                   2                   3
        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
header |V=2|P|    RC   |   PT=SR=200   |             length            |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                         SSRC of sender                        |
       +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
sender |              NTP timestamp, most significant word             |
info   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |             NTP timestamp, least significant word             |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                         RTP timestamp                         |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                     sender's packet count                     |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                      sender's octet count                     |
       +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+

}
{
 RTCP BYE Packet
       0                   1                   2                   3
       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |V=2|P|    SC   |   PT=BYE=203  |             length            |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      |                           SSRC/CSRC                           |
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
      :                              ...                              :
      +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
(opt) |     length    |               reason for leaving            ...
      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
}

 TRTCPHeader=record
  Version             : 0..2;
  Padding             : boolean;
  ReceptionReportCount: 0..31;
  PacketType          : byte;
  Length              : integer;
  SSRC                : cardinal;
  NTPTimestampHigh    : cardinal;
  NTPTimestampLow     : cardinal;
  RTPTimestamp        : cardinal;
  PacketCount         : cardinal;
  OctetCount          : cardinal;
 end;
{
       +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
report |                 SSRC_1 (SSRC of first source)                 |
block  +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
  1    | fraction lost |       cumulative number of packets lost       |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |           extended highest sequence number received           |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                      interarrival jitter                      |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                         last SR (LSR)                         |
       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
       |                   delay since last SR (DLSR)                  |
       +=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+
}
 TSenderReport=record
  SSRC                  : cardinal;
  FractionLost          : byte;
  CumulativePacketsLost : 0..$FFF;
  HighestSequence       : cardinal;
  InterarrivalJitter    : cardinal;
  LastSR                : cardinal;
  DelaySinceLastSR      : cardinal;
 end;

 TCustomRTCPClient=class(TCustomUDPClient)
 private
  fRTP:TCustomRTPClient;
  fRemote:TSockAddr;
  fRemoteSSRC :cardinal;
  fPacketCount:cardinal;
  fBytesCount :cardinal;
 {$IFDEF LISTEN}
  fActive:boolean;
  fThread:THandle;
  procedure Listen;
 {$ENDIF}
  procedure SendBye(Sock:TSockAddr; SSRC:cardinal);
 public
  constructor Create(RTP:TCustomRTPClient);
  procedure Start(Target:string; Port:word);
  function GetPayload(PayloadType:byte):string;
  procedure Enter;
  procedure Leave;
  property RTP:TCustomRTPClient read fRTP;
 end;

 TRTCPClient=class(TCustomRTCPClient)
 end;

implementation

{$IFDEF LISTEN}
function StartMe(Sender:TCustomRTCPClient):integer; stdcall;
begin
 Sender.Listen;
end;
{$ENDIF}

function RTCPHeaderEncode(const Header:TRTCPHeader):string;
var
 b:byte;
 w:word;
 l:cardinal;
begin
 SetLength(Result,28);
 with Header do begin
  b:=ReceptionReportCount+Version shl 6;
  if Padding   then b:=b or 32;
  Result[1]:=chr(b);
  Result[2]:=chr(PacketType);
  w:=Swap(Length div 4-1);
  move(w,Result[3],2);
  l:=SSRC; LSwap(l);
  move(l,Result[5],4);
  l:=NTPTimestampHigh; LSwap(l);
  Move(l,Result[9],4);
  l:=NTPTimestampLow; LSwap(l);
  Move(l,Result[13],4);
  l:=RTPTimestamp; LSwap(l);
  Move(l,Result[17],4);
  l:=PacketCount; LSwap(l);
  Move(l,Result[21],4);
  l:=OctetCount; LSwap(l);
  Move(l,Result[25],4);
 end;
end;

procedure RTCPHeaderDecode(const Data:string; var Header:TRTCPHeader);
var
 b:byte;
 w:word;
begin
 FillChar(Header,SizeOf(Header),0);
 if Length(Data)<28 then exit;
 with Header do begin
  b:=ord(Data[1]);
  Version:=b shr 6;
  Padding:=(b and 32)>0;
  ReceptionReportCount:=b and 31;
  PacketType:=ord(Data[2]);
  Move(Data[3],w,2);
  Length:=4*(Swap(w)+1);
  Move(Data[ 5],SSRC,4);
  LSwap(SSRC);
  Move(Data[ 9],NTPTimestampHigh,4);
  Move(Data[13],NTPTimestampLow,4);
  Move(Data[17],RTPTimestamp,4);
  LSwap(RTPTimestamp);
  Move(Data[21],PacketCount,4);
  LSwap(PacketCount);
  Move(Data[25],OctetCount,4);
  LSwap(OctetCount);
 end;
end;

constructor TCustomRTCPClient.Create(RTP:TCustomRTPClient);
begin
 inherited Create;
 fRTP:=RTP;
end;

procedure TCustomRTCPClient.Start(Target:string; Port:word);
{$IFDEF LISTEN}
var
 id:cardinal;
{$ENDIF} 
begin
// Start RTCP
 OpenSocket;
 BindSocket(fRTP.Port+1);
// Start RTP
 fRTP.Start(Target,Port);
// Remote
 fRemote:=fRTP.Remote;
 fRemote.sin_port:=htons(Port+1);
// Listen
{$IFDEF LISTEN}
 if fActive then exit;
 fActive:=True;
 fThread:=CreateThread(nil,0,@StartMe,pointer(Self),0,id);
{$ENDIF}
end;

function TCustomRTCPClient.GetPayload(PayloadType:byte):string;
var
 s :string;
 hd:TRTPHeader;
begin
 Result:='';
 s:=fRTP.Read;
 if length(s)<13 then exit;

// verify packet
 RTPHeaderDecode(s,hd);
 if (hd.Version<>2)or(hd.PayloadType<>PayloadType) then exit;

// verify source
 if (fRTP.From.sin_addr.S_addr<>fRTP.Remote.sin_addr.S_addr)
 or (fRTP.From.sin_port<>fRTP.Remote.sin_port) then begin
  exit;
 end;

// verify SSRC
 if fRemoteSSRC=0 then begin
  fRemoteSSRC:=hd.SSRC
 end else begin
  if hd.SSRC<>fRemoteSSRC then exit;
 end;

// ok !
 Result:=copy(s,13,MaxInt);
 inc(fPacketCount);
 inc(fBytesCount,Length(Result));
end;

{$IFDEF LISTEN}
procedure TCustomRTCPClient.Listen;
// this is for debug purpose, freephonie.net doesn't seems to send any RTCP packet !
var
 sa:TSockAddr;
 si:integer;
 ss:string;
 sl:integer;
 rc:integer;
 hd:TRTCPHeader;
begin
 try
  sl:=1500;
  SetLength(ss,sl);
  while fActive do begin
   FillChar(sa,SizeOf(sa),0);
   sa.sin_family:=AF_INET;
   si:=SizeOf(sa);
   rc:=recvfrom(fSocket,ss[1],sl,0,sa,si);
   if rc>0 then begin
    WriteLn('RTCP received from ',inet_ntoa(sa.sin_addr),':',ntohs(sa.sin_port),' (');
    RTCPHeaderDecode(ss,hd);
    WriteLn(' Version    : ',hd.Version);
    WriteLn(' PacketType : ',hd.PacketType);
    WriteLn(' Timestamp  : ',hd.RTPTimestamp);
    WriteLn(' PacketCount: ',hd.PacketCount);
    WriteLn(' BytesCount : ',hd.OctetCount);
    WriteLn(' Length     : ',rc);
    WriteLn(')');
   end;
  end;
 finally
  CloseHandle(fThread);
  fThread:=0;
 end;
end;
{$ENDIF}

procedure TCustomRTCPClient.Enter;
begin
 fRemoteSSRC:=0;
 fPacketCount:=0;
 fBytesCount:=0;
 fRTP.SendPayload(8,'');
end;

procedure TCustomRTCPClient.Leave;
var
 pk:string;
  w:word;
  l:cardinal;
begin
 SetLength(pk,12);
 pk[1]:=chr(2 shl 6); // version
 pk[2]:=#203; // BYE
 w:=Swap(Length(pk) div 4-1);
 move(w,pk[3],2);
 l:=fRemoteSSRC; LSwap(l);
 move(l,pk[5],4);
 pk[9]:=#3;
 pk[10]:='B';
 pk[11]:='Y';
 pk[12]:='E';
 for w:=0 to 5 do begin
  l:=fPacketCount;
  sendto(fSocket,pk[1],Length(pk),0,fRemote,SizeOf(fRemote));
  sleep(10);
  if l=fPacketCount then break;
 end;
{$IFDEF LISTEN}
 fActive:=False;
 WaitOrKill(fThread,10);
{$ENDIF}
end;

procedure TCustomRTCPClient.SendBye(Sock:TSockAddr; SSRC:cardinal);
var
 pk:string;
  w:word;
  l:cardinal;
begin
// WriteLn('SEND BYE to ',inet_ntoa(Sock.sin_addr),':',ntohs(Sock.sin_port));
 SetLength(pk,8);
 pk[1]:=chr(2 shl 6+1); // version
 pk[2]:=#203; // BYE
 w:=Swap(Length(pk) div 4-1);
 move(w,pk[3],2);
 l:=0; LSwap(l);
 move(l,pk[5],4);
(*
 pk[9]:=#3;
 pk[10]:='B';
 pk[11]:='Y';
 pk[12]:='E';
*)
 Sock.sin_port:=htons(ntohs(Sock.sin_port)+1);
 sendto(fSocket,pk[1],Length(pk),0,Sock,SizeOf(Sock));
end;

end.

